/*
 * Copyright (c) 2001 Network Appliance, Inc.
 * All rights reserved.
 */

#include "header.h"
#include "iserver.h"
#include "gbuf.h"
#include <stdio.h>
#include <assert.h>
#ifdef WINDOWS
#include <errno.h>
#else
#ifdef COMPILE_SOLARIS
#include <errno.h>
#else
#include <sys/errno.h>
#endif
#endif /* Unix/Windows */

/* Parse the ICAP header */
int
header_read_icap(iserver_conn_t *conn)
{
	int retval, icap_req_recognized=0;
	char buf[BUFLEN], *line;

	assert(conn);

	for (;;) {
		memset(buf, '\0', BUFLEN);

		retval = iserver_readline(conn,buf,BUFLEN);
		if (retval < 0) {
			if (errno == EPIPE) { /* 32 */ 
				iserver_logmsg(conn, "TCP connection closed\n");
			}
			conn->closed = 1;
			goto header_error;

		} else if (retval == 0)
			break;

		line = buf;

		if (line) {

			if (!MATCH(line, CRLF)) {
				
				if (cinfo.verbose)
					printf("Server-rhdr : %s", line);
				
				if (strstr(line, "res-hdr")) {
					conn->origin_res_hdr_exist = 1;
				} 
				if (strstr(line, "res-body")) {
					conn->origin_res_body_exist = 1;
				}
				if (MATCH(line, "OPTIONS")) {
					
					conn->icap_req_type = ICAP_OPTIONS;
					icap_req_recognized=1;

				} else if (MATCH(line, "REQMOD")) {
				
					conn->icap_req_type = ICAP_REQMOD;
					icap_req_recognized=1;

				} else if (MATCH(line, "RESPMOD")) {

					conn->icap_req_type = ICAP_RESPMOD;
					icap_req_recognized=1;

				} else if (MATCH(line, "LOG")) {
					
					conn->icap_req_type = ICAP_LOG;
					icap_req_recognized=1;

				} else 	if (MATCH(line, "Transfer-Encoding: chunked")) {
					
					iserver_logmsg(conn, "Transfer-Encoding field cannot exist for ICAP 1.0\n");
					goto header_error;

				} else if (MATCH(line, "Preview:")) {
					if (!sscanf(line, "Preview:%d", &conn->preview))
						conn->preview = 0;
					line = 0;
				} else if (MATCH(line, "TE:")) {
					/* NO-OP */
				} else {
					/* Dont want no headers besides 
					   one's we recognize */
					line = 0;
				}
			}
			
			/* add to request header */
			gbuf_concat(conn->req_hdr, line, retval);
			
			if ( line && MATCH(line, CRLF)) {
				if (cinfo.verbose)
					printf("Server-RXXX : %s", line);
				break;
			}
		}
	}

	if (icap_req_recognized)
		return 1;
	else {
		if (!retval)
			iserver_logmsg(conn, "conn closed by client\n");
		else
			iserver_logmsg(conn, "unrecognized ICAP request\n");
		
		conn->closed=1;
		return -1;
	}
	
  header_error:
	iserver_logmsg(conn, "error while parsing ICAP header\n");
	return -1;
}

/* parse HTTP client header */
int
header_read_client(iserver_conn_t *conn)
{
	char buf[BUFLEN];
	int retval;

	for (;;) {
		char *line;
		
		memset(buf, '\0', BUFLEN);
		retval = iserver_readline(conn,buf,BUFLEN);

		if (retval < 0) {
			if (errno == EPIPE) { /* 32 */ 
				iserver_logmsg(conn, "TCP connection closed\n");
			}
			conn->closed = 1;
			goto header_error;

		} else if (retval == 0)
			break;

		line = buf;

		if (line) {
			if (MATCH(line, "POST") && 
			    (conn->icap_req_type == ICAP_REQMOD)) {
				conn->icap_reqmod_post = 1;
			}
			
			gbuf_concat(conn->client_hdr, line, retval);
			
			if (cinfo.verbose)
				printf("Server-chdr : %s", line);
			
			if (MATCH(line, CRLF))
				break;
		}
	}
	return 1;

  header_error:
	iserver_logmsg(conn,"error while parsing HTTP client request hdr\n");
	return -1;
}

/* parse HTTP origin header for respmod */ 
int
header_read_origin(iserver_conn_t *conn)
{
	char buf[BUFLEN];
	int retval;

	for (;;) {
		char *line;
		
		memset(buf, '\0', BUFLEN);
		retval = iserver_readline(conn,buf,BUFLEN);
		if (retval < 0) {
			if (errno == EPIPE) { /* 32 */ 
				iserver_logmsg(conn,"TCP connection closed\n");
			}
			conn->closed = 1;
			goto header_error;

		} else if (retval == 0)
			break;

		line = buf;

		if (line) {
			if (!MATCH(line, CRLF)) {
				if (cinfo.verbose)
					printf("Server-ohdr : %s", line);

			} 
			if (MATCH(line, "HTTP/1.1")) {
				sscanf(line,
				       "HTTP/1.1 %d",
				       &conn->resp_code);
			} 
			if (MATCH(line, "Via")) { 
				/* insert the hostname in Via Field */
				if (*(line+strlen(line)-1) == '\n') {
					*(line+strlen(line)-1) = '\0';
					if (*(line+strlen(line)-1) == '\r') {
						*(line+strlen(line)-1) = '\0';
					}
				}
				memcpy(line+strlen(line), ", ", 2);
				assert(*hostname);
				memcpy(line+strlen(line), hostname, 
				       strlen(hostname));
				memcpy(line+strlen(line), "\r\n", 2);
				retval=strlen(line);
			}

			gbuf_concat(conn->origin_hdr, line, retval);

			if (line && MATCH(line, CRLF)) {
				if (cinfo.verbose)
					printf("Server-OXXX : %s", line);
				break;
			}
		}
	}
	return 1;

 header_error:
	iserver_logmsg(conn, "error while parsing origin header\n");
	return -1;
}
