/*
 * Copyright (c) 2001 Network Appliance, Inc.
 * All rights reserved.
 */

#include "iserver.h"
#include "threadpool.h"
#include "gbuf.h"
#include "header.h"
#include "stream.h"
#include "api.h"

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <sys/types.h>
#include <assert.h>
#include <signal.h>


#ifdef WINDOWS
#include <Windows.h>
#include <time.h>
#include <errno.h>
#else
#ifdef COMPILE_SOLARIS
#include <errno.h>
#else
#include <sys/errno.h>
#endif
#include <netinet/tcp.h> 
#include <sys/time.h>
#include <unistd.h>
#include <semaphore.h>
#endif

/*
 * GLOBALS
 */
#ifdef WINDOWS
static HANDLE time_mtx;
extern HANDLE available_sem; 
#else
static pthread_mutex_t time_mtx;
extern sem_t available_sem;
#endif

/* 
 * PROTOTYPES
 */
static int iserver_parse_body(iserver_conn_t *conn);
static iserver_conn_t *iserver_alloc_conn(void);
static void iserver_free_conn(iserver_conn_t *conn);
static void iserver_reset_conn(iserver_conn_t *conn);

/*
 * FUNCTIONS
 */
void
iserver_reg_api(void)
{
        FILE *fp;
	char line;

        /* set function pointers from api */
	iserver_options_response = api_options_response;
	iserver_log_request = api_log_request;
	iserver_preview_reqmod = api_preview_reqmod;
	iserver_preview_respmod = api_preview_respmod;
	iserver_modify_reqmod = api_modify_reqmod;
	iserver_modify_respmod = api_modify_respmod;
#ifdef WINDOWS	
	file_mtx = CreateMutex(NULL, FALSE, NULL);
#else
	pthread_mutex_init(&file_mtx, 0);
#endif
	options_str = (char*) malloc(2048*sizeof(char));
	memset(options_str, '\0', 2048*sizeof(char));
	if ((fp = fopen("options.conf", "r" )) == NULL) {
		printf("Error opening options.conf file");
		exit(-1);
	}
	for(line=(char)getc(fp); line!= EOF; line=(char)getc(fp)) {
		strncpy(options_str+strlen(options_str), &line, 1);
	}
	fclose(fp);
	
        /* get the hostname for via */
	hostname = (char*) malloc(2048*sizeof(char));
	memset(hostname, '\0', 2048*sizeof(char));
	assert(!gethostname(hostname, 2048));
}

void
iserver_await_connections(void)
{
	iserver_conn_t *lconn;
	int allow_addr_reuse=1, temp=0, no_delay=1;
#ifdef WINDOWS
	SOCKET srv_socket;
	WSADATA wsaData;
	if ((WSAStartup(MAKEWORD(2,2), &wsaData)) != 0)
		printf("Winsock init error\n");
	time_mtx = CreateMutex(NULL, FALSE, NULL);
	cinfo.mtx = CreateMutex(NULL, FALSE, NULL);
#else
	int srv_socket;
	pthread_mutex_init(&time_mtx, 0);
	pthread_mutex_init(&cinfo.mtx, 0);
#endif

	if (cinfo.modify_percent)
		iserver_logmsg(0, "ICAP content modifying server\n");
	else
		iserver_logmsg(0, "ICAP content echoing server\n");

	/* initialize threadpool */
	if (threadpool_init(cinfo.num_thr, 
			    (void *(*)(void *))iserver_connected) < 0) {
		iserver_logmsg(0, "Could not allocate threadpool\n");
		goto error;
	}

	iserver_logmsg(0, "Pre-created %d threads\n", cinfo.num_thr);
	
	/* NOTE: SIGINT does not work for windows */
	/* register the signal handler*/
	if ((*signal)(SIGINT, (void (*)(int))threadpool_signal_handler) == SIG_ERR) { 
		printf("registering ctrl-C sig handler"); 
		exit(-1); 
	} 
	
	/* ignore SIGPIPE to prevent iserver aborts in linux */
	if (sigignore(SIGPIPE) == -1) {
		printf("registering SIGPIPE sig handler");
		exit(-1);
	}
	/* setup listen socket */
	lconn = iserver_alloc_conn();
	if (!lconn)
		goto error;
	assert(lconn->content->blk_size);
	
	/* get socket */
	srv_socket = socket(AF_INET, SOCK_STREAM, 0);
#ifdef WINDOWS
	if (srv_socket == INVALID_SOCKET) {
		fprintf(stderr,"socket() failed with error %d\n",WSAGetLastError());
		goto error;
	}
#else
	if (srv_socket < 0) {
		printf("Socket Error");
		goto error;
	}
#endif
	/* 
	 * allow server-side addresses to be reused (don't have
	 * to wait for that annoying timeout)
	 */
	temp = setsockopt(srv_socket,
			  SOL_SOCKET,
			  SO_REUSEADDR,
			  (void *)&allow_addr_reuse,
			  sizeof(allow_addr_reuse));
#ifdef WINDOWS
	if (temp != 0) {
		printf("setsockopt error");
		goto error;
	}
#else
        if (temp < 0)  {
		printf("setsockopt error");
		goto error;
	}
#endif
	/* 
	 * turn off the Nagle's algorithm
	 */
	temp = setsockopt(srv_socket,
			  IPPROTO_TCP,
			  TCP_NODELAY,
			  (void *)&no_delay,
			  sizeof(no_delay));
#ifdef WINDOWS
	if (temp != 0) {
		printf("setsockopt error");
		goto error;
	}
#else
        if (temp < 0)  {
		printf("setsockopt error");
		goto error;
	}
#endif
	/* bind unnamed socket */
	lconn->addr.sin_family = AF_INET;
	lconn->addr.sin_addr.s_addr = htonl(INADDR_ANY);
	lconn->addr.sin_port = htons(cinfo.port);
	
	temp = bind(srv_socket, (struct sockaddr*)&lconn->addr, 
		    sizeof(lconn->addr)); 
#ifdef WINDOWS
	if (temp != 0) {
		printf("bind error");
		goto error;
	}
#else
	if (temp < 0) {
		printf("bind error");
		goto error;
	}
#endif 
	/* socket needs to listen for connections*/
        temp = listen(srv_socket, 10);

#ifdef WINDOWS
	if (temp != 0) {
		printf("listen error");
		goto error;
	}
#else
	if (temp < 0) {
		printf("listen error");
		goto error;
	}
#endif
	/* accept connections */	
	for (;;) {
		iserver_conn_t *cconn = iserver_alloc_conn();
		cconn->addr_len = sizeof(cconn->addr);
		
		/* wait for available thread */
#ifdef WINDOWS
		WaitForSingleObject(available_sem, INFINITE);
#else
		sem_wait(&available_sem);
#endif
		PTHREAD_MUTEX_LOCK(cinfo.mtx);
		cinfo.conn_total++;
		cconn->num = cinfo.conn_total;
	        PTHREAD_MUTEX_UNLOCK(cinfo.mtx);

		iserver_logmsg(0, 
			       "Waiting for connection %d on port %d", 
			       cconn->num, 
			       cinfo.port);
		cconn->skt = accept(srv_socket,
				    (struct sockaddr*)&cconn->addr,
				    &cconn->addr_len);
#ifdef WINDOWS
		if (cconn->skt == INVALID_SOCKET){
			printf("accept error");
			iserver_free_conn(cconn);
			goto error;
		}
#else
		if (cconn->skt < 0) { 
			printf("accept error");
			iserver_free_conn(cconn);
			goto error;
		}
#endif
		
	threadpool_dispatch( (void *)cconn );
}
	
  error:
	printf("\nError: Closing Server\n");
	threadpool_shutdown();
	printf("Closing Listen port.\n");
	if (lconn)
		iserver_free_conn(lconn);
#ifdef WINDOWS
	CloseHandle(time_mtx);	
	CloseHandle(cinfo.mtx);	
	if (srv_socket != INVALID_SOCKET)
		closesocket(srv_socket);
#else
	pthread_mutex_destroy(&time_mtx);
	pthread_mutex_destroy(&cinfo.mtx);
	if (srv_socket != -1)
		close(srv_socket);
#endif
}

static void
iserver_send_modified_response(iserver_conn_t* conn)
{
	ca_t* response;

	if (conn->icap_req_type == ICAP_REQMOD) {
		/* call reqmod api* with 
		   1) icap hdr + client hdr + entire post msg body)
		*/ 
		if (!conn->icap_reqmod_post)
			assert(conn->content->length == 0);
		if (conn->icap_reqmod_post && !conn->content->length)
			conn->content->length = -1;

		response = (*iserver_modify_reqmod)(conn->req_hdr->data, 
						    conn->client_hdr->data, 
						    conn->content->data,
						    conn->content->length);
	} 
	else { 
		/* call respmod api* with  
		   1) icap hdr + origin hdr + origin resp hdr +
		   entire msg body
		   
		   * - api not to care of preview length, just 
			   modify body (of any size) on basis of headers
		*/
		response = (*iserver_modify_respmod)(conn->req_hdr->data, 
						     conn->client_hdr->data, 
						     conn->origin_hdr->data,
						     conn->content->data,
						     conn->content->length);
	}
	
	assert(response);
	assert(response->beg);
	iserver_writereply(conn, response);
	free(response->beg);
	free(response);
} 	

/* entry point for all client-handling threads
 */
void
iserver_connected(iserver_conn_t *conn)
{
	int unsigned temp=0;
	
	/* don't need to lock, as add/sub is idempotent */
	cinfo.conn_active++	;

	while (1) {

		/* parse ICAP request header */
		if (header_read_icap(conn) < 0) {
			goto done;
		}
		
		if (conn->icap_req_type != ICAP_OPTIONS) {
			
			/* parse client request header */
			if (header_read_client(conn) < 0) {
				goto done;
			}
			
			if (conn->icap_req_type == ICAP_RESPMOD&&
				conn->origin_res_hdr_exist) {
				
				/* parse origin response headers */
				if (header_read_origin(conn) < 0) {
					goto done;
				}
			}
		
			if ( conn->icap_reqmod_post || 
			     (conn->icap_req_type == ICAP_RESPMOD
				     && conn->origin_res_body_exist)) {
			
				/* parse message body for preview (if any) */
				if (iserver_parse_body(conn) < 0) {
					goto done;
				}
			}
		}		
		else {
			/* send an options response */
		
			char* options_reply;
		
			/* FUNCPTR*/
			options_reply = (*iserver_options_response)(conn->icap_uri);
			assert(options_reply);
		
		        iserver_writeline(conn, options_reply);
			free(options_reply);
			goto done;
		}
		
		if (conn->icap_req_type == ICAP_LOG) {
			/* LOG request sent by client, no response to be sent */ 
		
			/* FUNCPTR*/
			(*iserver_log_request)(conn->req_hdr->data);
			goto done;
		}
	
		/* call appropriate API function to decide the 
		   preview response (if any) */	
		if (conn->preview >= 0) {
			char* response;
			
			if (conn->icap_req_type == ICAP_REQMOD) {
				/* call reqmod api with 
				   1) icap hdr + client hdr + post msg body
				   (max size = preview)
				   2) preview length
				*/
				if (!conn->icap_reqmod_post)
					assert( conn->content == NULL);
			
				response = (*iserver_preview_reqmod)(conn->req_hdr->data,
								     conn->client_hdr->data,
								     conn->content->data,
								     conn->content->length,
								     conn->preview);
			} 
			else { 
				/* call respmod api with  
				   1) icap hdr + client hdr + origin resp hdr +
				   msg body(max size = preview)
				   2) preview length
				*/
				response = (*iserver_preview_respmod)(conn->req_hdr->data,
								      conn->client_hdr->data,
								      conn->origin_hdr->data,
								      conn->content->data,
								      conn->content->length,
								      conn->preview);
			}
		
			/* 
			 *  1) send 100 Continue or 204 depending on
			 *     what api returned ( NULL - 100, 
			 *      non-NULL - write out data recvd, free mem and
			 *                 close connection
			 *  2) if (total msg body < preview) && (100 continue was sent)
			 *     send modified hdrs+content (from char* returned) by 
			 *     calling same api as the one used below
			 */
		
			if (response) {
				if (cinfo.verbose) {
					iserver_logmsg(conn, "Sending....\n");
					iserver_logmsg(conn, response);
				}
			
				temp = iserver_writeline(conn, response);
				free(response);
				if (temp<0)
					goto done;
			}
			else {
				if (cinfo.verbose && !conn->short_preview)
					iserver_logmsg(conn,
						       "Sending \"ICAP/1.0 100"
						       " Continue\\r\\n\\r\\n"
						       "message...\n"); 
			
				if (!conn->short_preview) {
					temp = iserver_writeline(conn, 
					      "ICAP/1.0 100 Continue\r\n\r\n");
					if (temp<0)
						goto done;
				}

				if ((conn->icap_reqmod_post || 
				     (conn->icap_req_type == ICAP_RESPMOD))&&
				    conn->short_preview) {
				
					iserver_send_modified_response(conn);
				
				}
		
			}

			/* total msg body < preview or 204 sent */ 
			if (response || conn->short_preview) 
				goto done;

		}
		
		/* get remaining msg body */
		if ( conn->icap_reqmod_post || 
		     (conn->icap_req_type == ICAP_RESPMOD
			     &&conn->origin_res_body_exist)) {
			
			if (read_body_chunked(conn) < 0)
				goto done;
		}

		/* send to api: initial+remaining msg body (if preview >= 0) 
		 * or total body (if preview <0)
		 */
		iserver_send_modified_response(conn);
		
	  done:
		if (CLOSECONN)
		  break;
		if (!conn->closed) {
			/* keep connection open for next request */
			iserver_reset_conn(conn);
			if (cinfo.verbose)
				iserver_logmsg(0, "Recycling- curr conn active: %d\n", cinfo.conn_active);

		}
		else {
			/* close connection */
			break;
		}
		
	}	
	
	fflush(0);
	
#ifdef WINDOWS
	closesocket(conn->skt);
#else
	close(conn->skt);
#endif

	cinfo.conn_active--;
	
	if (cinfo.verbose)
		iserver_logmsg(0, "Connections active: %d\n", cinfo.conn_active);
	
	iserver_free_conn(conn);
	
}

/* parse the body of the transmission
 */
static int
iserver_parse_body(iserver_conn_t *conn)
{
	char *line;
		
	assert(conn);

	if (conn->icap_req_type == ICAP_RESPMOD && conn->origin_res_hdr_exist) {

		if (!conn->origin_hdr->length) {
			iserver_logmsg(conn, "TCP connection closed");
			conn->closed = 1;
			return -1;
		}
		
		line = conn->origin_hdr->data;
		
		if ( strstr(line, "Content-Type:") ) {
			if ( strstr(line, "Content-Type: text/html") ||
			     strstr(line, "Content-Type: text/text") ) {
				if (cinfo.verbose)
					iserver_logmsg(conn, "HTML content found\n");
				conn->html_content = 1;
				
			} else {
				if (cinfo.verbose)
					iserver_logmsg(conn, 
						       "Unknown content type "
						       "(probably an image)\n");
				conn->html_content = 0;
			}
			
		} else  {
			/* no content type specified; assume text/html */
			conn->html_content = 1;
		}
	}

	return read_preview_chunked(conn);
}

/* write (null-terminated) string str to the connection represented by conn */
int
iserver_writeline(iserver_conn_t *conn, const char *str)
{
	unsigned int bytes = 0;

	assert(conn && str);
	
	bytes = WRITE(conn->skt, str, strlen(str));
	
	if (errno == SIGPIPE) {
		iserver_logmsg(conn, "TCP connection closed");
		conn->closed=1;
		return -1;
	}

	if (bytes < 0) {
		printf("socket write error");
		return bytes;
	}

	assert(bytes == strlen(str));
	return bytes;
}

/* write ca_t to the connection represented by conn */
int
iserver_writereply(iserver_conn_t *conn, const ca_t *str)
{
	int written = 0;
	
	assert(conn && str);
	
	/* Sanity checks */
	assert(strstr(str->beg, "ICAP/1.0") || 
	       strstr(str->beg, "HTTP"));
	assert(strstr(str->beg, "ISTAG"));

	while (written < (str->len-1)) {
		written += WRITE(conn->skt, str->beg, (str->len-1));
		if (errno == EPIPE) {
			iserver_logmsg(conn, "TCP connection closed\n");
			conn->closed=1;
			return -1;
		}
		if (written < 0) {
			printf("socket write error");
			return written;
		}
	}

	if (cinfo.verbose) {
		printf("bytes written at iserver_writereply: %d", written);
		printf("Reply written:\n%s", str->beg);
	}
	return written;
}

/* 
 * read a line from the connection represented by conn into buf, 
 * return num bytes read
 */
int
iserver_readline(iserver_conn_t *conn, char *buf, int buflen)
{
	char c;
	int i, retval;
	
	assert(conn && buf);

	i = 0;
	for (;;) {

	  if ( (retval = READ(conn->skt, &c, 1)) <= 0)
	    break;
		
		buf[i++] = c;
		if (c == '\n')
			break;

		if (i >= buflen)
			break;
	}
	if (retval < 0) {
		printf("socket read error");
		return retval;
	}

	if (i >= buflen) {
		printf("\n *** Buffer size exceeded in iserver_readline ***\n");
		assert(0);
	}
	buf[i] = 0;

	return i;
}

/* allocate an iserver_conn_t, set defaults, and return it */
static iserver_conn_t *
iserver_alloc_conn(void)
{
	iserver_conn_t *c;

	c = (iserver_conn_t *)malloc(sizeof(iserver_conn_t));
	memset(c, '0', sizeof(iserver_conn_t));

	if (!c) {
		printf("malloc error");
		exit(-1);
	}
	
	c->chunked = 0;
	c->chunk_size = 0;
	c->content_len = -1;
	c->resp_code = 0;
	c->html_content = 0;
	c->closed = 0;
	c->preview = -1;
	c->short_preview = 0;
	c->icap_req_type = -1;
	c->icap_uri = 0;
	c->icap_reqmod_post = 0;
	c->origin_res_hdr_exist = 0;
	c->origin_res_body_exist = 0;
	
	if (! (c->content = gbuf_alloc(CONTENT_BLK_SIZE)) ) {
		printf("gbuf allocation");
		return 0;
	}
	if (! (c->req_hdr = gbuf_alloc(CONTENT_BLK_SIZE)) ) {
		printf("gbuf allocation");
		return 0;
	}
	if (! (c->client_hdr = gbuf_alloc(CONTENT_BLK_SIZE)) ) {
		printf("gbuf allocation");
		return 0;
	}
	if (! (c->origin_hdr = gbuf_alloc(CONTENT_BLK_SIZE)) ) {
		printf("gbuf allocation");
		return 0;
	}

	return c;
}

/* free an iserver_conn_t */
static void
iserver_free_conn(iserver_conn_t *conn)
{
	assert (conn);
	
	if (conn->icap_uri)
	        free(conn->icap_uri);
	
	if (conn->req_hdr)
		gbuf_free(conn->req_hdr);

	if (conn->client_hdr)
		gbuf_free(conn->client_hdr);

	if (conn->origin_hdr)
		gbuf_free(conn->origin_hdr);

	if (conn->content)
		gbuf_free(conn->content);

	free(conn);
}

/* reset an iserver_conn_t
 */
static void
iserver_reset_conn(iserver_conn_t *conn)
{
	assert(conn);

	conn->chunked = 0;
	conn->chunk_size = 0;
	conn->content_len = -1;
	conn->resp_code = 0;
	conn->html_content = 0;
	conn->closed = 0;
	conn->preview = -1;
	conn->short_preview = 0;
	conn->icap_req_type = -1;
	conn->icap_uri = 0;
	conn->icap_reqmod_post = 0;
	
	if (conn->icap_uri)
	        free(conn->icap_uri);
	conn->icap_uri = 0;
	
	gbuf_zero(conn->req_hdr);
	gbuf_zero(conn->client_hdr);
	gbuf_zero(conn->origin_hdr);
	gbuf_zero(conn->content);

	return;
}

/* print (date and?) timestamped log messages to stderr */
void
iserver_logmsg(iserver_conn_t *conn, const char *fmt,...)
{
	struct tm *timeptr;
	va_list ap;
	time_t t;
	
	assert(fmt);
	
	PTHREAD_MUTEX_LOCK(time_mtx);
	
	time(&t);
	timeptr = localtime(&t);
	
	fprintf(stderr, "[");
	if (timeptr->tm_hour < 10)
		fprintf(stderr, "0");
	fprintf(stderr, "%d:", timeptr->tm_hour);
	
	if (timeptr->tm_min < 10)
		fprintf(stderr, "0");
	fprintf(stderr, "%d:", timeptr->tm_min);
	
	if (timeptr->tm_sec < 10)
		fprintf(stderr, "0");
	fprintf(stderr, "%d", timeptr->tm_sec);
	
        PTHREAD_MUTEX_UNLOCK(time_mtx);
	
	if (conn)
		fprintf(stderr, " (c%d)", conn->num);
	
	fprintf(stderr, "] ");
	
	va_start(ap,fmt);
	vfprintf(stderr, fmt, ap);
	
	va_end(ap);
	
	return;
}
