/*
 * Copyright (c) 2001 Network Appliance, Inc.
 * All rights reserved.
 */

#ifndef _ISERVER_H_
#define _ISERVER_H_

#include <stdarg.h>

#undef WINDOWS

#ifdef WINDOWS
#include <Winsock2.h>
#else
#include <pthread.h>
#include <sys/socket.h>
#include <semaphore.h>
#ifdef COMPILE_SOLARIS
#include <netinet/in.h>
#else 
#include <arpa/inet.h>
#endif
#endif /* Windows/Unix */

#define CONTENT_BLK_SIZE		4096
#define READ_BLK_SIZE			1024
#define BUFLEN				4096
#define HDRLEN				4096
#define CRLF				"\r\n"
#define DEFAULT_PORT			1344
#define DEFAULT_NUM_THREADS             150
#define DEFAULT_REQMOD_CACHE_PERCENT    0
#define DEFAULT_MODIFY_PERCENT          100

#define MATCH(a,b)			(!strncmp((a),(b), strlen((b))))
#define MIN(a,b)			(( (b) > (a) ) ? (a) : (b))
#define MAX(a,b)			(( (b) > (a) ) ? (b) : (a))

/* close connection after dealing with requests (debugging) 
   or leave open for persistence */
#define CLOSECONN                       0

/* Platform specific sys calls */
#ifdef WINDOWS
#define PTHREAD_MUTEX_LOCK(m)    WaitForSingleObject(m, INFINITE)      
#define PTHREAD_MUTEX_UNLOCK(m)  ReleaseMutex(m)
#define READ(conn, buf, sz)      recv(conn, buf, sz, 0)
#define WRITE(conn, buf, sz)     send(conn, buf, sz, 0)
#else
#define PTHREAD_MUTEX_LOCK(m)    pthread_mutex_lock(&m)
#define PTHREAD_MUTEX_UNLOCK(m)  pthread_mutex_unlock(&m)
#define READ(conn, buf, sz)      read(conn, buf, sz)   
#define WRITE(conn, buf, sz)     write(conn, buf, sz)
#endif

enum icap_request_t {
        ICAP_OPTIONS = 0,
        ICAP_REQMOD,
	ICAP_RESPMOD,
	ICAP_LOG
};

struct gbuf;

typedef struct iserver_conn {

	/* thread & connection stats info */
        int 			num;
#ifdef WINDOWS
        HANDLE  		thr;	
	SOCKET 			skt;
#else
        pthread_t               thr;
        int                     skt;
#endif 
        struct sockaddr_in	addr;
	int 			addr_len;

	/* HTML content info */
	int 			chunked;
	int			chunk_size;
	int 			preview;
        int                     short_preview;
        int 			content_len;
	int			resp_code;
	int			html_content;
	int 			closed;
       
        /* ICAP request info */
        char*                   icap_uri;
        enum icap_request_t     icap_req_type;       
	int                     icap_reqmod_post;
        int                     origin_res_hdr_exist;
        int                     origin_res_body_exist;

        /* the webpage itself */
	struct gbuf 		*req_hdr;
	struct gbuf	    	*client_hdr;
	struct gbuf		*origin_hdr;
	struct gbuf	      	*content;
	
} iserver_conn_t;

typedef struct iserver_info {
#ifdef WINDOWS
        HANDLE 	                mtx;
#else
        pthread_mutex_t         mtx;
#endif
  	int                     num_thr;
	int 			port;
	int 			verbose;
	int                     trailer_support;
        int                     reqmod_cache_percent;
        int                     modify_percent;
	
        int 			conn_total;
	int 			conn_active;
} iserver_info_t;

typedef struct ca{
        char* beg;
        int len;
} ca_t;

/*
 * GLOBALS
 */
iserver_info_t cinfo;
char *options_str;
char *hostname;

/*
 * PROTOTYPES
 */
extern void 
iserver_await_connections(void);
	
extern void 
iserver_connected(struct iserver_conn *conn);

extern int 
iserver_readline(struct iserver_conn *conn, 
		 char *buf, 
		 int buflen);

extern int 
iserver_writeline(struct iserver_conn *conn, 
		  const char *str);
extern int 
iserver_writereply(struct iserver_conn *conn, 
		   const ca_t *str);

extern void 
iserver_logmsg(struct iserver_conn *conn, 
	       const char *fmt,...);

/* register the API's with iserver */
extern void 
iserver_reg_api(void);

/* 
 *Function pointers for API 
 */

/* obtain an options response for a particular ICAP service */
char* (*iserver_options_response)(const char* icap_uri);

/* process the LOG request sent by client */
void  (*iserver_log_request)(const char* icap_hdr); 

/* obtain preview response for reqmod to send back to client */
char*   (*iserver_preview_reqmod)(const char* icap_hdr, 
				  const char* client_hdr, 
				  const char* post_msg_body,
				  int post_msg_body_len,
				  int preview_size);

/* obtain preview response for respmod to send back to client */
char*   (*iserver_preview_respmod)(const char* icap_hdr, 
				   const char* client_hdr, 
				   const char* origin_resp_hdr,
				   const char* origin_resp_body,
				   int origin_resp_body_len,
				   int preview_size);

/* obtain a modified request (and body) for reqmod to send back to 
   client */
ca_t* (*iserver_modify_reqmod)(const char* icap_hdr, 
			       const char* client_hdr, 
			       const char* post_msg_body,
			       int post_msg_body_len);

/* obtain modified request and body for respmod to send back to 
   client */
ca_t* (*iserver_modify_respmod)(const char* icap_hdr, 
				const char* client_hdr, 
				const char* origin_resp_hdr,
				const char* origin_resp_body,
				int origin_resp_body_len);

#endif // _ISERVER_H_
