/*
 * Copyright (c) 2001 Network Appliance, Inc.
 * All rights reserved.
 */

#include "stream.h"
#include "iserver.h"
#include "gbuf.h"
#include <assert.h>
#include <stdio.h>

#define STREAM_READ_FAILED -1

int
read_preview_chunked(iserver_conn_t *conn)
{
	int nread, bytes, read_len;
	char *p;
	char buf[BUFLEN];
	char *dbuf = 0;

	assert(conn);
	
	nread = 0;
	while (nread <= conn->preview) {
		
		memset(buf, '\0', BUFLEN);
		if ( (bytes = iserver_readline(conn, buf, BUFLEN)) <= 0 ) {
			iserver_logmsg(conn,
				       "error reading chunked\n");
			nread = STREAM_READ_FAILED;
			conn->closed = 1;
			goto done;
		}

		/* chop carriage-return linefeed */
		assert( p = (char *)strstr(buf, CRLF) );
		if (p) {
			*p = 0;
			*(p+1) = 0;
		}

		/* convert hex chunk size to decimal */
		if (!sscanf(buf, "%x", &conn->chunk_size)) {
			iserver_logmsg(conn, "error converting chunk size\n");
			nread = STREAM_READ_FAILED;
			goto done;
		}

		if (conn->chunk_size == 0) { 
		        /* last chunk indicator for preview*/
			/* check for "0; ieof\r\n" 
			 * to decide if this is end of the body also
			 */
			if (!strcmp(buf, "0; ieof"))
				conn->short_preview = 1; 
			/* read extra \r\n as per correction to ICAP10 spec */
			iserver_readline(conn, buf, BUFLEN);
			goto done;
		}

		read_len = (conn->chunk_size > conn->preview) ?
			conn->preview : conn->chunk_size;
	
		dbuf = (char *)malloc(read_len);
		if (!dbuf) {
			iserver_logmsg(conn, "error allocating dbuf\n");
			goto done;
		}
		memset(dbuf, '\0', read_len);

		while (bytes = READ(conn->skt, dbuf, read_len)) {

			gbuf_concat(conn->content, dbuf, bytes);
			nread += bytes;
			read_len -= bytes;
			
			if (conn->html_content) {
				if (cinfo.verbose)
					printf("\nCTNT (Size %d): %s\n", 
					       conn->content->length,
					       conn->content->data);
			}

			if (!read_len) {
				break;
			}

			memset(dbuf, '\0', read_len);
			
			if (bytes == 0) {
				iserver_logmsg(conn, "chunk2:  TCP closed!\n");
				conn->closed = 1;
				nread = STREAM_READ_FAILED;
				goto done;
			}
		}

		free(dbuf); 
		dbuf = 0;

		if (nread == conn->chunk_size) {
			/* eat trailing \r\n  */
			iserver_readline(conn, buf, BUFLEN);
			
		}
	}

	if (cinfo.verbose)
		iserver_logmsg(conn, "Preview size: %d, read: %d", conn->preview, nread);

 done:
	if (dbuf)
		free(dbuf);
	return nread;
}

int
read_body_chunked(iserver_conn_t *conn)
{
	int nread = 0;
	int temp = 0;
	int rsize, bytes;
	char *p=0, *dbuf=0, buf[BUFLEN];
	
	if (conn->preview > 0) {
	
		assert(conn);
		
		if (conn->content->length < conn->chunk_size ) {
			conn->chunk_size -= conn->content->length;
			goto chunk_partial;
		}
	}
	
	for (;;) {
		/* eat chunk size*/
		iserver_readline(conn, buf, BUFLEN);
		if (!buf[0]) { 
                        /* TCP closed */
			iserver_logmsg(conn, "blank line close (surprise)\n");
			conn->closed = 1;
			nread = STREAM_READ_FAILED;
			goto done;
		}

		/* chop carriage-return linefeed */
		assert( p = (char *)strstr(buf, CRLF) );
		if (p) {
			*p = 0;
			*(p+1) = 0;
		}

		/* convert hex chunk size to decimal */
		if (!sscanf(buf, "%x", &conn->chunk_size)) {
			iserver_logmsg(conn, "error converting chunk size\n");
			nread = STREAM_READ_FAILED;
			goto done;
		}

		if (conn->chunk_size == 0) { 
		        /* last chunk indicator for msg body */
			/* received "0; ieof\r\n"  or "0\r\n" */

			/* read extra \r\n as per correction to ICAP10 spec */
			iserver_readline(conn, buf, BUFLEN);
			goto done;
		}

		/* eat chunk body */
	chunk_partial:
		rsize = conn->chunk_size;

		dbuf = (char *)malloc(conn->chunk_size);
		if (!dbuf) {
			perror("malloc error");
			nread = STREAM_READ_FAILED;
			goto done;
		}
		memset(dbuf, '\0', conn->chunk_size);		
		
		while (bytes = READ(conn->skt, dbuf, rsize)) {

			nread += bytes;

			if (bytes == 0) { 
                                /* TCP closed */
				if (cinfo.verbose)
					iserver_logmsg(conn, "chunk2: TCP closed (partial chunk)!\n");
				conn->closed = 1;
				nread = STREAM_READ_FAILED;
				goto done;
			}
			
			if(conn->html_content) {
				if (cinfo.verbose)
					printf("CTNT:  %s\n",dbuf);
			}

			gbuf_concat(conn->content, dbuf, bytes);
			
			rsize-= bytes;
			
			if (!rsize){
				/* eat trailing \n */
				iserver_readline(conn, buf, BUFLEN);
				break;
			}
			
			memset(dbuf, '\0', conn->chunk_size);
		}
		
		free(dbuf);
		dbuf=0;
	}
 done: 
	if (dbuf)
		free(dbuf);
	return nread;
}
